from typing import Generator, List, Union
import numpy as np
from ...sq_packet import SQPacket, SQMetadata
from ...sq_queue import SQQueue
from ...sq_event import SQEvent
from ..sq_helper import SQHelper
from .sq_packet_info import SQPacketInfo
from .sq_generic_packet import SQGenericPacket
from ...sq_time_base import SQTimeBase


def register(helper_factory):
    helper_factory.register(name="default", factory=SQDefaultHelper)


class SQDefaultHelper(SQHelper):

    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.no_pkts_mean = data.get('no_pkts_mean', 10)
        self.no_pkts_sd = data.get('no_pkts_sd', 2)
        self.pkt_size_mean = data.get('pkt_size_mean', 1000)
        self.pkt_size_sd = data.get('pkt_size_sd', 2000)
        self.classes = data.get('classes', ['A'])
        self.priorities = data.get('priorities', [1, 10])
        self.pkt_id = 0
        self.q_idx = 0

    def generate_packets(self) -> Generator[List[SQPacket], None, None]:
        pkts = []
        no_of_pkts = int(np.abs(np.random.normal(self.no_pkts_mean, self.no_pkts_sd, None)))
        pkt_sizes = [int(x) for x in np.abs(np.random.normal(self.pkt_size_mean, self.pkt_size_sd, no_of_pkts))]
        pkt_classes = np.random.choice(self.classes, no_of_pkts)
        pkt_priorities = np.random.randint(self.priorities[0], self.priorities[1], no_of_pkts)
        pkt_info: SQPacketInfo = SQPacketInfo(no_of_pkts, pkt_sizes, pkt_classes, pkt_priorities)
        for p in range(pkt_info.no_of_pkts):
            pkt = SQGenericPacket(id=self.pkt_id,
                                  size=pkt_info.pkt_sizes[p],
                                  class_name=pkt_info.pkt_classes[p],
                                  priority=pkt_info.pkt_priorities[p],
                                  timestamp=SQTimeBase.get_current_sim_time())
            self.pkt_id += 1
            pkts.append(pkt)
        yield pkts

    def get_processing_cycles(self, pkt: SQPacket) -> int:
        np.random.seed(pkt.size)
        return np.random.randint(1, 10)

    def process_packet(self, pkt: SQPacket, tick: int) -> Union[SQMetadata, None]:
        self.owner.logger.info(f'Processing packet {pkt} at tick {tick}')
        if tick < self.get_processing_cycles(pkt):
            return None
        return SQMetadata(name='tick_count', owner=self.owner.name, value=tick)

    def filter_packet(self, pkt: SQPacket) -> bool:
        return True

    def process_data(self, data: SQEvent, tick: int):
        self.owner.logger.info(f'Consuming Metadata {data}')

    def select_input_queue(self) -> Union[SQQueue, None]:
        if hasattr(self.owner, "input_qs"):
            q = self.owner.input_qs[self.q_idx]
            self.q_idx = (self.q_idx + 1) % len(self.owner.input_qs)
            return q
        else:
            return None

    def select_output_queue(self, pkt) -> Union[SQQueue, None]:
        if hasattr(self.owner, "output_qs"):
            q = self.owner.output_qs[self.q_idx]
            self.q_idx = (self.q_idx + 1) % len(self.owner.output_qs)
            return q
        else:
            return None



from .sq_generator_state import SQPktGeneratorState
from ...sq_event import SQEvent


class SQPktGeneratorQueuingState(SQPktGeneratorState):
    def process_packet(self, evt: SQEvent):
        if len(self.owner.packets)>0:
            self.owner.output_q.push(self.owner.packets[self.owner.tick])
            self.owner.generated_pkts += 1
            self.owner.total_pkts += 1
            self.owner.logger.info(f'Packet {self.owner.packets[self.owner.tick]} Ready for Queuing')
            self.owner.tick += 1
        if self.owner.tick >= len(self.owner.packets):
            self.owner.packets = []
            self.owner.tick = 0
            self.owner.set_state(self.factory.create_state(name='GENERATING', owner=self.owner))
        self.owner.finish_indication()

    def get_state_name(self):
        return f'QUEUING'



import copy
from typing import Union

from ..sq_event import SQEvent
from ..sq_logger import SQLogger
from ..sq_object import SQObject
from ..sq_queue import SQQueue
from ..sq_plugin import SQHelper


class SQMux(SQObject):
    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.input_qs = data.get('input_qs', [])
        if len(self.input_qs) < 2:
            raise ValueError('At least two input_qs should be provided')
        self.output_q = data.get('output_q', None)
        if self.output_q is None:
            raise ValueError('output_q should be provided')
        self.clk = data.get('clk', None)
        if self.clk is not None:
            self.clk.control_flow(self)
        else:
            raise ValueError('Clock not provided')
        for p in self.input_qs:
            if p is not None:
                if not isinstance(p, SQQueue):
                    raise ValueError(f'queues should contain  SQQueue object ,'
                                     f' got {type(p)} instead.')

            else:
                raise ValueError('Null Queue Provided')
        if self.helper is None:
            raise ValueError('Helper not provided')
        self.current_port = None

    def process_packet(self, evt):
        super().process_packet(evt)
        if evt.owner is self.clk:
            self.current_port = self.helper.select_input_queue()
            if self.current_port is not None:
                curr_pkt = self.current_port.pop()
                if curr_pkt is not None:
                    self.output_q.push(copy.copy(curr_pkt))
                    self.finish_indication()
        else:
            if evt.name != f'{self.name}_start':
                self.logger.error(f'Ignoring Events other than Clock Events {evt}')

    def process_data(self, evt: SQEvent):
        super().process_data(evt)
        self.helper.process_data(evt,self.tick)



import unittest
from unittest.mock import MagicMock
from dataclasses import dataclass
from ...src.pysysq import *
from ...src.pysysq.sq_base.sq_plugin.default.sq_generic_packet import SQGenericPacket


class TestSQDemux(unittest.TestCase):

    def setUp(self):
        self.factory = SQFactory()
        self.tx_q1 = self.factory.create('SQQueue', {'name': 'tx_q1', 'capacity': 10})
        self.tx_q2 = self.factory.create('SQQueue', {'name': 'tx_q2', 'capacity': 10})
        self.tx_q3 = self.factory.create('SQQueue', {'name': 'tx_q3', 'capacity': 10})
        self.tx_qs = [self.tx_q1, self.tx_q2, self.tx_q3]
        self.input_q = self.factory.create('SQQueue', {'name': 'input_q', 'capacity': 10})
        self.clk = self.factory.create('SQClock', {'name': 'clk', 'clk_divider': 1})

        self.demux = self.factory.create(obj_type='SQDemux',
                                         data={
                                             'name': 'demux',
                                             'output_qs': self.tx_qs,
                                             'input_q': self.input_q,
                                             'clk': self.clk}
                                         )
        self.simulator = None

    def run_sim_loops(self, no_of_sim_loops: int):
        self.simulator = self.factory.create(obj_type='SQSimulator',
                                             data={
                                                 'name': 'simulator',
                                                 'max_sim_time': no_of_sim_loops,
                                                 'time_step': 0.1,
                                                 'children': [
                                                     self.clk,
                                                     self.demux,
                                                     self.tx_q1,
                                                     self.tx_q2,
                                                     self.tx_q3,
                                                     self.input_q
                                                 ]
                                             }
                                             )
        self.simulator.init()
        self.simulator.start()

    def test_demux_queue_selection(self):
        # Arrange

        # push a packet to each rx queue
        for i in range(3):
            print(f'pushing packet = {i * 10} to input_q')
            self.input_q.push(SQGenericPacket(id=i * 10))

        # Act
        self.run_sim_loops(no_of_sim_loops=4)

        # Assert
        for i in range(3):
            id = self.tx_qs[i].pop().id
            print(f'popped packet = {id} from {self.tx_qs[i].name}')
            self.assertEqual(id, i * 10)





from .sq_time_base import SQTimeBase
from .sq_object import SQObject
from .sq_event import *
from .sq_clock import *
from .sq_factory import *
from .sq_filter import *
from .sq_logger import *
from .sq_merger import *
from .sq_mux_demux import *
from .sq_packet import *
from .sq_pkt_processor import *
from .sq_pkt_gen import *
from .sq_queue import *
from .sq_simulator import *
from .sq_splitter import *
from .sq_statistics import *
from .sq_sim_setup_generator import *



from .sq_pkt_processor_state import SQPktProcState
from ...sq_event import SQEvent


class SQPktProcStateIdle(SQPktProcState):

    def process_packet(self, evt: SQEvent):
        self.owner.curr_pkt = self.owner.input_queue.pop()
        self.owner.start_tick = self.owner.tick
        if self.owner.curr_pkt is None:
            self.owner.logger.info(f'{self.owner.name} No Packet to Process')
        else:
            self.owner.pkt_size_sum += self.owner.curr_pkt.size
            self.owner.pkt_size_average = self.owner.pkt_size_sum / (self.owner.no_of_processed_pkts + 1)
            self.owner.processing_time = self.owner.helper.get_processing_cycles(self.owner.curr_pkt)
            metadata = self.owner.helper.process_packet(self.owner.curr_pkt, self.owner.tick)
            self.owner.update_progress()
            if metadata is not None:
                self.owner.data_indication(data=metadata)

            self.owner.logger.info(f'{self.owner.name} Start Processing Packet '
                                   f'{self.owner.curr_pkt} Expected processing time '
                                   f'{self.owner.processing_time} current Tick {self.owner.tick}')
            self.owner.set_state(self.factory.create_state('PROCESSING', owner=self.owner))

    def get_state_name(self):
        return f'IDLE'



import unittest
from unittest.mock import MagicMock
from dataclasses import dataclass
from ...src.pysysq import *
from ...src.pysysq.sq_base.sq_plugin.default.sq_generic_packet import SQGenericPacket


class TestSQDemux(unittest.TestCase):

    def setUp(self):
        self.factory = SQFactory()
        self.rx_q1 = self.factory.create('SQQueue', {'name': 'rx_q1', 'capacity': 10})
        self.rx_q2 = self.factory.create('SQQueue', {'name': 'rx_q2', 'capacity': 10})
        self.rx_q3 = self.factory.create('SQQueue', {'name': 'rx_q3', 'capacity': 10})
        self.rx_qs = [self.rx_q1, self.rx_q2, self.rx_q3]
        self.output_q = self.factory.create('SQQueue', {'name': 'input_q', 'capacity': 10})
        self.clk = self.factory.create('SQClock', {'name': 'clk', 'clk_divider': 1})

        self.mux = self.factory.create(obj_type='SQMux',
                                       data={
                                           'name': 'mux',
                                           'output_q': self.output_q,
                                           'input_qs': self.rx_qs,
                                           'clk': self.clk}
                                       )
        self.simulator = None

    def run_sim_loops(self, no_of_sim_loops: int):
        self.simulator = self.factory.create(obj_type='SQSimulator',
                                             data={
                                                 'name': 'simulator',
                                                 'max_sim_time': no_of_sim_loops,
                                                 'time_step': 0.1,
                                                 'children': [
                                                     self.clk,
                                                     self.mux,
                                                     self.rx_q1,
                                                     self.rx_q2,
                                                     self.rx_q3,
                                                     self.output_q
                                                 ]
                                             }
                                             )
        self.simulator.init()
        self.simulator.start()

    def test_mux_queue_selection(self):
        # Arrange

        self.rx_q1.push(SQGenericPacket(id=0))
        self.rx_q2.push(SQGenericPacket(id=10))
        self.rx_q3.push(SQGenericPacket(id=20))

        # Act
        self.run_sim_loops(no_of_sim_loops=4)

        # Assert
        for i in range(3):
            id = self.output_q.pop().id
            print(f'popped packet = {id}')
            self.assertEqual(id, i * 10)



from abc import ABC, abstractmethod
from .sq_sim_data_model import (SQSimDataModel, DataFlow)


class SQSimDataGenStrategy(ABC):
    def __init__(self, context):
        self.context = context

    @abstractmethod
    def generate(self, data: dict) -> SQSimDataModel:
        pass

    def get_array_repr(self, array: list):
        return '[' + ','.join([f'self.{x}'.lower() for x in array]) + ']'

    def append_helper_params(self, data: dict, object_data: dict):
        helper_params = data.get('helper_params', {})
        for k, v in helper_params.items():
            object_data[k] = v


    def get_data_flows(self, data: dict):
        data_flows = []
        if 'data_flow' not in data:
            return data_flows
        for d in data['data_flow']:
            data_flows.append(DataFlow(data=d['data'], destination=d['destination']))
        return data_flows


class SQClockDataGenStrategy(SQSimDataGenStrategy):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate(self, data: dict) -> SQSimDataModel:
        obj_data = dict(
            name="\"" + data['name'] + "\"",
            clk_divider=data['clk_divider']
        )
        self.append_helper_params(data, obj_data)
        clk_model = SQSimDataModel(
            name=data['name'],
            type=data['type'],
            comment=data['description'],
            data_flows=self.get_data_flows(data),
            plot=data['plot'],
            sq_object_data=obj_data,
            children=[],
        )
        return clk_model


class SQSimulatorDataGenStrategy(SQSimDataGenStrategy):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate(self, data: dict) -> SQSimDataModel:
        children = [self.context.generate(x) for x in data['children']]
        children_names = self.get_array_repr([x.name for x in children])

        print(children_names)
        obj_data = dict(
            name="\"" + data['name'] + "\"",
            max_sim_time=data['max_sim_time'],
            time_step=data['time_step'],
            children=children_names,
        )
        self.append_helper_params(data, obj_data)
        return SQSimDataModel(
            name=data['name'],
            type=data['type'],
            comment=data['description'],
            data_flows=self.get_data_flows(data),
            plot=data['plot'],
            sq_object_data=obj_data,
            children=children,
        )


class SQSISODataGenStrategy(SQSimDataGenStrategy):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate(self, data: dict) -> SQSimDataModel:
        obj_data = dict(
            name="\"" + data['name'] + "\"",
            clk='self.' + str(data['clk']).lower(),
            input_q='self.' + str(data['input_q']).lower(),
            output_q='self.' + str(data['output_q']).lower(),
            helper="\"" + data.get('helper', 'default') + "\""
        )
        self.append_helper_params(data, obj_data)
        return SQSimDataModel(
            name=data['name'],
            type=data['type'],
            comment=data['description'],
            data_flows=self.get_data_flows(data),
            plot=data['plot'],
            sq_object_data=obj_data,
            children=[],
        )


class SQSIMODataGenStrategy(SQSimDataGenStrategy):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate(self, data: dict) -> SQSimDataModel:
        output_qs = self.get_array_repr(data['output_qs'])
        obj_data = dict(
            name="\"" + data['name'] + "\"",
            clk='self.' + str(data['clk']).lower(),
            input_q='self.' + str(data['input_q']).lower(),
            output_qs=output_qs,
            helper="\"" + data.get('helper', 'default') + "\""
        )
        self.append_helper_params(data, obj_data)
        return SQSimDataModel(
            name=data['name'],
            type=data['type'],
            comment=data['description'],
            data_flows=self.get_data_flows(data),
            plot=data['plot'],
            sq_object_data=obj_data,
            children=[],
        )


class SQMISODataGenStrategy(SQSimDataGenStrategy):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate(self, data: dict) -> SQSimDataModel:
        input_qs = self.get_array_repr(data['input_qs'])
        obj_data = dict(
            name="\"" + data['name'] + "\"",
            clk='self.' + str(data['clk']).lower(),
            input_qs=input_qs,
            output_q='self.' + str(data['output_q']).lower(),
            helper="\"" + data.get('helper', 'default') + "\""
        )
        self.append_helper_params(data, obj_data)
        return SQSimDataModel(
            name=data['name'],
            type=data['type'],
            comment=data['description'],
            data_flows=self.get_data_flows(data),
            plot=data['plot'],
            sq_object_data=obj_data,
            children=[],
        )


class SQSINODataGenStrategy(SQSimDataGenStrategy):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate(self, data: dict) -> SQSimDataModel:
        obj_data = dict(
            name="\"" + data['name'] + "\"",
            clk='self.' + str(data['clk']).lower(),
            input_q='self.' + str(data['input_q']).lower(),
            helper="\"" + data.get('helper', 'default') + "\""
        )
        self.append_helper_params(data, obj_data)
        return SQSimDataModel(
            name=data['name'],
            type=data['type'],
            comment=data['description'],
            data_flows=self.get_data_flows(data),
            plot=data['plot'],
            sq_object_data=obj_data,
            children=[],

        )


class SQNISODataGenStrategy(SQSimDataGenStrategy):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate(self, data: dict) -> SQSimDataModel:
        obj_data = dict(
            name="\"" + data['name'] + "\"",
            clk='self.' + str(data['clk']).lower(),
            output_q='self.' + str(data['output_q']).lower(),
            helper="\"" + data.get('helper', 'default') + "\""
        )
        self.append_helper_params(data, obj_data)
        return SQSimDataModel(
            name=data['name'],
            type=data['type'],
            comment=data['description'],
            data_flows=self.get_data_flows(data),
            plot=data['plot'],
            sq_object_data=obj_data,
            children=[],
        )


class SQQueueDataGenStrategy(SQSimDataGenStrategy):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate(self, data: dict) -> SQSimDataModel:
        obj_data = dict(
            name="\"" + data['name'] + "\"",
            capacity=data['capacity'],
        )
        self.append_helper_params(data, obj_data)
        qs = SQSimDataModel(
            name=data['name'],
            type=data['type'],
            comment=data['description'],
            data_flows=self.get_data_flows(data),
            children=[],
            plot=data['plot'],
            sq_object_data=obj_data,
        )
        return qs



import matplotlib.pyplot as plt
import pandas as pd
from ..sq_logger import SQLogger


class SQPlotter:
    def __init__(self, name: str, objs: [], **kwargs):
        self.name = name
        self.objs = objs
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.show_plot = kwargs.get('show_plot', False)
        self.output_file = kwargs.get('output_file', 'Statistics.png')

    def plot(self):
        plt.figure()
        for obj in self.objs:
            self.logger.debug(f'Plotting {obj.name}')
            properties = obj.read_statistics().get_all_property_names(obj.name)
            for property_name in properties:
                self.plot_property(property_name=property_name, obj=obj)
        plt.legend()
        plt.xlabel('Simulation Time')
        plt.ylabel('Property Value')
        plt.title(f'SQObject Properties')
        plt.savefig(self.output_file)
        if self.show_plot:
            plt.show()

    def generate_excel(self, filename):

        # Create a pandas ExcelWriter object
        writer = pd.ExcelWriter(filename, engine='xlsxwriter')

        for obj in self.objs:
            properties = obj.read_statistics().get_all_property_names(obj.name)
            for property_name in properties:
                property_values = obj.read_statistics().get_property(name=property_name, owner=obj.name)
                x_values = [entry.sim_time for entry in property_values]
                y_values = [entry.value for entry in property_values]

                # Create a DataFrame for this property
                df = pd.DataFrame({
                    'Simulation Time': x_values,
                    'Property Value': y_values
                })
                sheet_name = f'{obj.name}_{property_name}'
                if len(sheet_name) > 31:
                    sheet_name = sheet_name[:31]
                # Write the DataFrame to a worksheet named after the object and property
                df.to_excel(writer, sheet_name=f'{sheet_name}', index=False)

        # Save the Excel file
        writer.close()

    def plot_property(self, property_name: str, obj):
        self.logger.debug(f'Plotting {obj.name} [{property_name}]')
        property_values = obj.read_statistics().get_property(name=property_name, owner=obj.name)
        x_values = [entry.sim_time for entry in property_values]
        y_values = [entry.value for entry in property_values]
        plt.plot(x_values, y_values, label=f'{obj.name}[{property_name}]')



import copy
from typing import List
from ..sq_object import SQObject
from ..sq_queue import SQQueue
from ..sq_logger import SQLogger
from ..sq_clock import SQClock


class SQSplitter(SQObject):
    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.output_qs = data.get('output_qs', [])
        if len(self.output_qs) < 2:
            raise ValueError('At least two output_qs should be provided')
        self.input_q = data.get('input_q', None)
        if self.input_q is None:
            raise ValueError('input_q should be provided')
        self.clk = data.get('clk', None)

        if self.clk is not None:
            self.clk.control_flow(self)
        else:
            raise ValueError('Clock not provided')
        for p in self.output_qs:
            if not isinstance(p, SQQueue):
                raise ValueError(f'rx_q should be a SQQueue object , got {type(p)} instead.')

    def process_packet(self, evt):
        super().process_packet(evt)
        if evt.owner is self.clk:
            pkt = self.input_q.pop()
            for q in self.output_qs:
                if pkt is not None:
                    self.logger.info(f'Pushing Packet {evt.data} to Queue {q.name}')
                    q.push(copy.copy(pkt))
            self.finish_indication()
        else:
            if evt.name != f'{self.name}_start':
                self.logger.error(f'Ignoring Events other than Clock Event {evt}')



from .sq_pkt_processor_state import SQPktProcState
from ...sq_event import SQEvent


class SQPktProcStateComplete(SQPktProcState):
    def process_packet(self, evt: SQEvent):
        self.owner.no_of_processed_pkts += 1
        self.owner.total_processing_time += self.owner.processing_time
        self.owner.avg_processing_time = self.owner.total_processing_time / self.owner.no_of_processed_pkts
        self.owner.load = self.owner.total_processing_time / self.owner.tick * 100
        metadata = self.owner.helper.process_packet(self.owner.curr_pkt, self.owner.tick)
        self.owner.update_progress()
        if metadata is not None:
            self.owner.data_indication(data=metadata)

        self.owner.logger.info(f'{self.owner.name} Packet {self.owner.curr_pkt} Processing Complete after ticks '
                               f'{self.owner.tick}')
        self.owner.output_queue.push(self.owner.curr_pkt)
        self.owner.finish_indication()
        self.owner.set_state(self.factory.create_state('IDLE', owner=self.owner))

    def get_state_name(self):
        return f'COMPLETE'



import time


class SQTimeBase:
    current_sim_time = 0
    current_host_time = 0

    @classmethod
    def get_current_sim_time(cls):
        return cls.current_sim_time

    @classmethod
    def get_current_host_time(cls):
        return time.time()

    @classmethod
    def update_current_sim_time(cls):
        cls.current_sim_time += 1

    @classmethod
    def reset_current_sim_time(cls):
        cls.current_sim_time = 0



from abc import ABC, abstractmethod

from ...sq_event import SQEvent


class SQPktGeneratorState(ABC):
    def __init__(self, owner, factory):
        self.owner = owner
        self.factory = factory

    @abstractmethod
    def process_packet(self, evt: SQEvent):
        pass

    @abstractmethod
    def get_state_name(self):
        pass

    def __repr__(self):
        return f'{self.get_state_name()}'



from typing import Union
from .sq_pkt_processor_state import SQPktProcState
from .sq_pkt_proc_idle import SQPktProcStateIdle
from .sq_pkt_proc_processing import SQPktProcStateProcessing
from .sq_pkt_proc_complete import SQPktProcStateComplete


class SQStateFactory:

    def create_state(self, name: str, owner) -> Union[SQPktProcState, None]:
        state = None
        if name == "IDLE":
            state = SQPktProcStateIdle(owner=owner, factory=self)
        elif name == "PROCESSING":
            state = SQPktProcStateProcessing(owner=owner, factory=self)
        elif name == "COMPLETE":
            state = SQPktProcStateComplete(owner=owner, factory=self)
        else:
            state = None
        return state



from typing import List
from ..sq_object import SQObject
from ..sq_queue import SQQueue
from ..sq_logger import SQLogger
from ..sq_clock import SQClock


class SQMerger(SQObject):
    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.input_qs = data.get('input_qs', [])
        self.output_q = data.get('output_q', None)
        if self.output_q is None:
            raise ValueError('output_q should be provided')
        self.clk = data.get('clk', None)
        if self.clk is not None:
            self.clk.control_flow(self)
        if len(self.input_qs) < 2:
            raise ValueError('At least two input_qs should be provided')
        if self.output_q is None:
            raise ValueError('tx_queue should be provided')
        for p in self.input_qs:
            if not isinstance(p, SQQueue):
                raise ValueError(f'input_q should be a SQQueue object , got {type(p)} instead.')
        if not isinstance(self.output_q, SQQueue):
            raise ValueError(f'output_q should be a SQQueue object , got {type(self.output_q)} instead.')

    def process_packet(self, evt):
        super().process_packet(evt)
        if evt.owner is self.clk:
            for q in self.input_qs:
                if q.peek() is not None:
                    self.output_q.push(q.pop())
            self.finish_indication()
        else:
            if evt.name != f'{self.name}_start':
                self.logger.error(f'Ignoring other Events other than Clock Events {evt}')



from typing import Union
from .sq_generator_state import SQPktGeneratorState
from .sq_pkt_generator_gen_state import SQPktGeneratorGenState
from .sq_pkt_generator_queuing_state import SQPktGeneratorQueuingState


class SQGeneratorStateFactory:
    def create_state(self, name: str, owner) -> Union[SQPktGeneratorState, None]:
        state = None
        if name == "GENERATING":
            state = SQPktGeneratorGenState(owner=owner, factory=self)
        elif name == "QUEUING":
            state = SQPktGeneratorQueuingState(owner=owner, factory=self)
        else:
            state = None
        return state



import copy
from typing import Union
from ..sq_clock import SQClock
from ..sq_object import SQObject
from ..sq_logger import SQLogger
from ..sq_queue import SQQueue
from ..sq_plugin import SQHelper


class SQFilter(SQObject):
    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)

        self.input_q: Union[SQQueue, None] = data.get('input_q', None)
        if self.input_q is None:
            raise ValueError('Input Queue not provided')
        self.output_q: Union[SQQueue, None] = data.get('output_q', None)
        if self.output_q is None:
            raise ValueError('Output Queue not provided')
        self.clk: Union[SQClock, None] = data.get('clk', None)
        if self.clk is None:
            raise ValueError('Clock not provided')
        self.clk.control_flow(self)
        if self.helper is None:
            raise ValueError('Helper not provided')

    def process_packet(self, evt):
        super().process_packet(evt)
        if evt.owner is self.clk:
            pkt = self.input_q.pop()
            if pkt is not None:
                if self.helper.filter_packet(pkt):
                    self.output_q.push(copy.copy(pkt))
                    self.finish_indication(data=evt.data)
        else:
            if evt.name != f'{self.name}_start':
                self.logger.error(f'Ignoring Events other than Clock Events {evt}')



from typing import Union
from importlib import import_module
import inspect
from .sq_plugin import SQPlugin


class SQPluginLoader:
    def __init__(self, helper_factory):
        self._helper_factory = helper_factory

    @staticmethod
    def _import(name: str) -> SQPlugin:
        return import_module(name)  # type: ignore

    def load_plugin(self, plugin_name: str) -> Union[SQPlugin, None]:
        try:
            module = self._import(name=plugin_name)
            module.register(self._helper_factory)

        except Exception as e:
            print(f"Error loading plugin {plugin_name}: {e}")
            return None



from dataclasses import dataclass
from typing import List


@dataclass
class DataFlow:
    data: str
    destination: str


@dataclass
class SQSimDataModel:
    name: str
    type: str
    comment: str
    data_flows: List[DataFlow]
    children: List["SQSimDataModel"]
    plot: bool
    sq_object_data: dict





from typing import List, Union
import copy

from .sq_event import SQEvent
from ..sq_logger import SQLogger
from ..sq_time_base import SQTimeBase


class SQEventQueue:
    def __init__(self, name: str = ""):
        self.name = name
        self.queue: List[SQEvent] = []
        self.logger = SQLogger(self.__class__.__name__, name)

    def set_log_level(self, level):
        self.logger.set_level(level)

    def schedule(self, _event: SQEvent):

        self.queue.append(copy.copy(_event))

    def pop_next_event(self) -> Union[SQEvent, None]:
        if len(self.queue) > 0:
            evt = self.get_next_event()
            if evt is not None:
                self.logger.debug(f'pop event  {evt.name} ')
                self.queue.remove(evt)
            return evt
        else:
            return None

    def get_next_event(self):
        evt = None
        if len(self.queue) > 0:
            current_time = SQTimeBase.get_current_sim_time()
            possible_events = [k for k in self.queue if k.scheduled_tick <= current_time]
            if len(possible_events) > 0:
                evt = min(possible_events, key=lambda x: x.scheduled_tick)
        return evt



from typing import Callable, Dict

from .sq_sim_data_model import SQSimDataModel


class SQCodeGenModel:
    def __init__(self, data: SQSimDataModel) -> None:
        self.model = data
        self.imports = ['from pysysq import *']
        self.queues = []
        self.clocks = []
        self.sim_objects = []
        self.simulator = None
        self.plugins = []
        self.plot_objects = []
        self.name = ""
        self.parser: Dict[str, Callable[[SQSimDataModel], None]] = {
            'SQSimulator': self._parse_simulator,
            'SQQueue': self._parse_queues,
            'SQClock': self._parse_clocks,
            'SQFilter': self._parse_sim_object,
            'SQMerger': self._parse_sim_object,
            'SQMux': self._parse_sim_object,
            'SQDemux': self._parse_sim_object,
            'SQPacketGenerator': self._parse_sim_object,
            'SQPktProcessor': self._parse_sim_object,
            'SQPktSink': self._parse_sim_object,
            'SQSplitter': self._parse_sim_object
        }
        self._parse_simulator(self.model)

    def _parse_simulator(self, data: SQSimDataModel) -> None:
        for child in data.children:
            self.parser[child.type](data=child)
        self.simulator = data
        self.name = self.simulator.sq_object_data['name']

    def _parse_queues(self, data: SQSimDataModel) -> None:
        self.queues.append(data)
        if data.plot:
            self.plot_objects.append(data.name)

    def _parse_clocks(self, data: SQSimDataModel) -> None:
        self.clocks.append(data)
        if data.plot:
            self.plot_objects.append(data.name)

    def _parse_sim_object(self, data: SQSimDataModel) -> None:
        self.sim_objects.append(data)
        if data.plot:
            self.plot_objects.append(data.name)

    def generate_params(self, data: Dict[str, str]) -> str:
        ret_val = ', '.join([f'{k}={v}' for k, v in data.items()])
        return ret_val



from typing import List, Union
from ..sq_logger import SQLogger
from ..sq_queue import SQQueue
from ..sq_packet import SQPacket
from ..sq_event import SQEvent


class SQSingleQueue(SQQueue):
    def push(self, pkt: SQPacket):
        super().push(pkt)
        if len(self.queue) < self.capacity:
            self.queue.append(pkt)
            self.logger.info(f'Packet Queued {pkt}')
            self.pending_pkt_count = len(self.queue)
            self.total_enqueued_pkt_count += 1
            self.finish_indication()
        else:
            self.logger.warning(f' Queue Full , Dropping Packet {pkt}')
            self.pending_pkt_count = len(self.queue)
            self.dropped_pkt_count += 1

    def peek(self):
        if len(self.queue) == 0:
            return None
        return self.queue[0]

    def is_empty(self) -> bool:
        return len(self.queue) == 0

    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.queue: List[SQPacket] = []
        self.capacity = data.get('capacity', 10)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.dropped_pkt_count = 0
        self.pending_pkt_count = 0
        self.total_enqueued_pkt_count = 0
        self.register_property('dropped_pkt_count')
        self.register_property('pending_pkt_count')
        self.register_property('total_enqueued_pkt_count')

    def pop(self, **kwargs) -> Union[SQPacket, None]:
        super().pop()
        if len(self.queue) == 0:
            return None
        return self.queue.pop(0)

    def process_packet(self, evt: SQEvent):
        super().process_packet(evt)
        if evt.owner is not self:
            self.push(evt.data)
        else:
            if evt.name != f'{self.name}_start':
                self.logger.error(f'Ignoring Self Event {evt}')



from .sq_generator_state import SQPktGeneratorState
from .sq_pkt_generator_gen_state import SQPktGeneratorGenState
from .sq_pkt_generator_queuing_state import SQPktGeneratorQueuingState
from .sq_generator_state_factory import SQGeneratorStateFactory



from .sq_pkt_processor_state import SQPktProcState
from ...sq_event import SQEvent


class SQPktProcStateProcessing(SQPktProcState):
    def process_packet(self, evt: SQEvent):
        metadata = self.owner.helper.process_packet(self.owner.curr_pkt, self.owner.tick)
        if metadata is not None:
            self.owner.data_indication(data=metadata)
        self.owner.update_progress()
        self.owner.logger.info(f'{self.owner.name} Continue Processing Packet '
                               f'{self.owner.curr_pkt} Time {self.owner.tick}')
        if self.owner.tick - self.owner.start_tick >= self.owner.processing_time - 2:
            self.owner.set_state(self.factory.create_state('COMPLETE', owner=self.owner))

    def get_state_name(self):
        return f'PROCESSING'



import json
import os
import shutil

from .sq_sim_data_gen_ctx import SQSimDataGenCtx
from jinja2 import FileSystemLoader, Environment
from .sq_code_gen_model import SQCodeGenModel

_default_json_file = os.path.join(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config'), "input.json")


class SQSimSetupGen:
    def __init__(self, json_file: str = _default_json_file):
        with open(json_file, 'r') as file:
            self.data = json.load(file)
        self.gen_ctx = SQSimDataGenCtx()
        self.template_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
        file_loader = FileSystemLoader(self.template_folder)
        self.env = Environment(loader=file_loader, trim_blocks=True, lstrip_blocks=True)

    def generate(self, output_folder: str):
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)
        os.makedirs(output_folder)
        simulators = self.data['Simulators']
        for simulator in simulators:
            data = self.gen_ctx.generate(simulator)
            code_data_model = SQCodeGenModel(data)
            template = self.env.get_template('sim_setup.py.j2')
            output = template.render(model=code_data_model)
            self.create_file(data=output, output_folder=output_folder, file=f'{simulator["name"].lower()}_setup.py')
        print("Done")

    @staticmethod
    def create_file(data, output_folder, file):
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        file_name = os.path.join(output_folder, file)
        f = open(file_name, "w")
        f.write(data)
        f.write('\n')
        f.close()



import time
from ..sq_object import SQObject
from ..sq_event import SQEvent
from ..sq_event import SQEventManager
from ..sq_logger import SQLogger
from ..sq_time_base import SQTimeBase


class SQSimulator(SQObject):

    def start(self):
        super().start()
        self.process_packet(self.start_evt)

    def process_packet(self, evt: SQEvent):
        while SQTimeBase.get_current_sim_time() < self.max_sim_time:
            self.event_manager.run()
            SQTimeBase.update_current_sim_time()
            self.tick += 1
            self.logger.debug(f'Simulator Tick {self.tick}')
            time.sleep(self.time_step)
        self.event_manager.run()
        self.deinit()

    def __init__(self, data: dict[str, any]):
        """
        Constructor for the SQSimulator
        :param name: Name of the Simulator
        :param event_mgr: Event Manager to be used
        :param kwargs: Optional parameters
            max_sim_time: Maximum Simulation Time
            time_step: Time Step (in seconds)
        """
        super().__init__(data)
        SQTimeBase.reset_current_sim_time()
        self.max_sim_time: int = data.get('max_sim_time', 100)
        self.time_step: float = data.get('time_step', 0.1)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.self_starting = True



from typing import Generator, List, Union

from abc import abstractmethod
from ..sq_packet import SQPacket, SQMetadata

from ..sq_event import SQEvent


class SQHelper:

    def __init__(self, data: dict[str, any]):
        self.owner = data.get('owner', None)
        self.data = data

    def set_owner(self, owner):
        self.owner = owner

    @abstractmethod
    def generate_packets(self) -> Generator[List[SQPacket], None, None]:
        pass

    @abstractmethod
    def get_processing_cycles(self, pkt: SQPacket) -> int:
        pass

    @abstractmethod
    def process_packet(self, pkt: SQPacket, tick: int) -> Union[SQMetadata, None]:
        pass

    @abstractmethod
    def filter_packet(self, pkt: SQPacket) -> bool:
        pass

    @abstractmethod
    def process_data(self, data: SQEvent, tick: int):
        pass

    @abstractmethod
    def select_input_queue(self) :
        pass

    @abstractmethod
    def select_output_queue(self, pkt):
        pass



from typing import Union

from ..sq_object import SQObject
from ..sq_clock import SQClock
from ..sq_queue import SQQueue
from ..sq_plugin import SQHelper
from ..sq_logger import SQLogger
from .states import SQStateFactory
from .states import SQPktProcState
from ..sq_packet import SQMetadata
from ..sq_event import SQEvent


class SQPktProcessor(SQObject):
    def __init__(self, data: dict[str, any]):
        """
        Constructor for the SQPktProcessor
        :param name: Name of the Packet Processor
        :param event_mgr: Event Manager to be used
        :param kwargs: Dictionary of optional parameters
            clk: Clock to be used for timing
            input_queue: Queue from which the packets are received
            output_queue: Queue to which the processed packets are sent
            helper: Helper to be used for processing the packets.
            the helper should be a subclass of SQPktProcessorHelper
        """
        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.clk: Union[SQClock, None] = data.get('clk', None)
        if self.clk is None:
            raise ValueError('Clock not provided')
        self.input_queue: Union[SQQueue, None] = data.get('input_q', None)
        if self.input_queue is None:
            raise ValueError('Input Queue should be provided')
        self.output_queue: Union[SQQueue, None] = data.get('output_q', None)
        if self.output_queue is None:
            raise ValueError('Output Queue should be provided')
        if not isinstance(self.input_queue, SQQueue):
            raise ValueError(f'Input Queue should be a SQ Queue')

        if self.helper is None:
            raise ValueError('Helper should be provided')
        self._state_factory = SQStateFactory()
        self._state = self._state_factory.create_state(name='IDLE', owner=self)
        self.processing_time = 0
        self.curr_pkt = None
        self.start_tick = 0
        self.no_of_processed_pkts = 0
        self.pkt_size_average = 0
        self.pkt_size_sum = 0
        self.avg_processing_time = 0
        self.total_processing_time = 0
        self.load = 0
        # self.register_property('no_of_processed_pkts')
        # self.register_property('pkt_size_average')
        # self.register_property('avg_processing_time')
        # self.register_property('state')
        self.register_property('load')
        if self.clk is not None:
            self.clk.control_flow(self)
        else:
            raise ValueError('Clock not provided')

    def set_state(self, state: SQPktProcState):
        self._state = state

    def process_packet(self, evt):
        super().process_packet(evt)
        if evt.owner is self.clk:
            self.tick += 1
            self._state.process_packet(evt)
        else:
            self.logger.warning(f'{self.name} Ignoring Event {evt.data}')

    def update_progress(self):
        progress = (self.tick - self.start_tick) / self.processing_time * 100
        progress_metadata = SQMetadata(name='progress', owner=self.name, value=progress)
        self.data_indication(data=progress_metadata)

    def process_data(self, evt: SQEvent):
        super().process_data(evt)
        self.helper.process_data(evt.data, self.tick)



import copy
from typing import List, Union
from ..sq_object import SQObject
from ..sq_queue import SQQueue
from ..sq_clock import SQClock
from ..sq_logger import SQLogger
from ..sq_plugin import SQHelper
from ..sq_event import SQEvent


class SQDemux(SQObject):
    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.output_qs: List[SQQueue] = data.get('output_qs', [])
        self.input_q: Union[SQQueue, None] = data.get('input_q', None)
        self.clk: Union[SQClock, None] = data.get('clk', None)
        if self.clk is not None:
            self.clk.control_flow(self)
        else:
            raise ValueError('input_q should be provided')
        if self.input_q is None:
            raise ValueError('input_q should be provided')
        if len(self.output_qs) < 2:
            raise ValueError('At least two output queue should be provided')

        for p in self.output_qs:
            if not isinstance(p, SQQueue):
                raise TypeError(f'queues should contain SQQueue, got {type(p)} instead.')
        if not isinstance(self.input_q, SQQueue):
            raise TypeError(f'input_q should be a SQQueue object , got {type(self.input_q)} instead.')
        if self.helper is None:
            raise ValueError('Helper not provided')
        self.current_port = 0

    def process_packet(self, evt):
        super().process_packet(evt)
        if evt.owner is self.clk:
            curr_pkt = self.input_q.pop()
            self.current_port = self.helper.select_output_queue(curr_pkt)
            if self.current_port is not None:
                if curr_pkt is not None:
                    self.current_port.push(copy.copy(curr_pkt))
                    self.finish_indication()
            else:
                self.logger.error(f'No Queue Selected for {evt.data}')
        else:
            if evt.name != f'{self.name}_start':
                self.logger.error(f'Ignoring Events other than clk events {evt}')

    def process_data(self, evt: SQEvent):
        super().process_data(evt)
        self.helper.process_data(evt,self.tick)



from typing import Callable, Any
from ..sq_plugin import SQHelper
from ..sq_plugin.default import SQDefaultHelper


class SQHelperFactory:
    def __init__(self):
        self.factory_map: dict[str, Callable[..., SQHelper]] = {'default': SQDefaultHelper}

    def register(self, name, factory):
        self.factory_map[name] = factory

    def create(self, name, data: dict[str, Any]) -> SQHelper:
        return self.factory_map[name](data)



import argparse
from .sq_base import SQSimSetupGen

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate simulation setup.')
    parser.add_argument('--json_file', type=str, required=True, help='Path to the input JSON file.')
    parser.add_argument('--output_folder', type=str, required=True, help='Path to the output folder.')
    args = parser.parse_args()
    sim = SQSimSetupGen(json_file=args.json_file)
    sim.generate(output_folder=args.output_folder)


import logging

from ..sq_time_base import SQTimeBase


class SQLogger:
    def __init__(self, logger_name: str, owner: str):
        self.logger = logging.getLogger(logger_name)
        self.name = owner

    def set_level(self, level):
        self.logger.setLevel(level)

    def debug(self, message):
        self.logger.debug(f'{self.name}::{SQTimeBase.get_current_sim_time()}:: {message}')

    def info(self, message):
        self.logger.info(f'{self.name}::{SQTimeBase.get_current_sim_time()}:: {message}')

    def warning(self, message):
        self.logger.warning(f'{self.name}::{SQTimeBase.get_current_sim_time()}:: {message}')

    def error(self, message):
        self.logger.error(f'{self.name}::{SQTimeBase.get_current_sim_time()}:: {message}')

    def critical(self, message):
        self.logger.critical(f'{self.name}::{SQTimeBase.get_current_sim_time()}:: {message}')



from dataclasses import dataclass
from typing import List
from ..sq_time_base import SQTimeBase


@dataclass
class SQStatisticsEntry:
    sim_time: int
    value: int
    name: str
    owner: str


class SQStatistics:
    def __init__(self):
        self._statistics: List[SQStatisticsEntry] = []

    def add(self, sq_property: str, value: int, owner: str):
        entry: SQStatisticsEntry = SQStatisticsEntry(sim_time=SQTimeBase.get_current_sim_time(),
                                                     value=value,
                                                     name=sq_property,
                                                     owner=owner)
        self._statistics.append(entry)

    def get_property(self, name, owner: str):
        return list(filter(lambda x: x.name == name and x.owner == owner, self._statistics))

    def get_all_properties(self, owner: str) -> dict:
        return {x.name: list(filter(lambda y: y.name == x.name and x.owner == owner, self._statistics)) for x in
                self._statistics}

    def get_all_property_names(self, owner: str):
        return list(set([x.name for x in self._statistics if x.owner == owner]))



from abc import ABC, abstractmethod
from ..sq_object import SQObject
from ..sq_packet import SQPacket


class SQQueue(SQObject):

    def pop(self, **kwargs):
        self.collect_statistics()

    @abstractmethod
    def is_empty(self) -> bool:
        pass

    @abstractmethod
    def peek(self):
        pass

    def push(self, pkt: SQPacket):
        self.collect_statistics()



from enum import Enum
from typing import Callable


class EventType(Enum):
    START_EVT = 0
    PROCESS_EVT = 1
    METADATA_EVT = 2
    TICK_EVENT = 3


class SQEvent:
    def __init__(self, _name: str, owner):
        self.name = _name
        self.owner = owner
        self.sim_queuing_timestamp = 0
        self.sim_processing_timestamp = 0
        self.host_timestamp = None
        self.scheduled_tick = 0
        self.data = None
        self.type:EventType = EventType.PROCESS_EVT
        self.actions = []

    def add_handler(self, action: Callable):
        if action not in self.actions:
            self.actions.append(action)

    def remove_handler(self, action: Callable):
        self.actions.remove(action)

    def has_handlers(self):
        return len(self.actions) > 0

    def __repr__(self):
        return self.name



from .sq_sim_data_gen_strategy import SQSimulatorDataGenStrategy, \
    SQClockDataGenStrategy, SQSISODataGenStrategy, SQMISODataGenStrategy, \
    SQSIMODataGenStrategy, SQNISODataGenStrategy, SQSINODataGenStrategy, SQQueueDataGenStrategy


class SQSimDataGenCtx:
    def __init__(self):
        self.str = None
        self.strategy_map = {
            'SQSimulator': SQSimulatorDataGenStrategy(ctx=self),
            'SQClock': SQClockDataGenStrategy(ctx=self),
            'SQFilter': SQSISODataGenStrategy(ctx=self),
            'SQMerger': SQMISODataGenStrategy(ctx=self),
            'SQMux': SQMISODataGenStrategy(ctx=self),
            'SQDemux': SQSIMODataGenStrategy(ctx=self),
            'SQPacketGenerator': SQNISODataGenStrategy(ctx=self),
            'SQPktProcessor': SQSISODataGenStrategy(ctx=self),
            'SQPktSink': SQSINODataGenStrategy(ctx=self),
            'SQQueue': SQQueueDataGenStrategy(ctx=self),
            'SQSplitter': SQSIMODataGenStrategy(ctx=self)

        }

        self.type = None

    @property
    def strategy(self):
        return self.str

    @strategy.setter
    def strategy(self, strategy):
        self.str = strategy

    def generate(self, data: dict):
        self.type = data['type']
        if self.type in self.strategy_map:
            self.strategy = self.strategy_map[self.type]
        else:
            raise ValueError(f'Invalid Type {self.type}')
        return self.strategy.generate(data)



from typing import Callable, Any, List, Union
from ..sq_object import SQObject
from ..sq_clock import SQClock
from ..sq_filter import SQFilter
from ..sq_merger import SQMerger
from ..sq_mux_demux import SQMux, SQDemux
from ..sq_pkt_gen import SQPacketGenerator
from ..sq_pkt_processor import SQPktProcessor
from ..sq_simulator import SQSimulator
from ..sq_splitter import SQSplitter
from ..sq_pkt_sink import SQPktSink
from ..sq_plugin import SQPluginLoader
from .sq_helper_factory import SQHelperFactory
from ..sq_queue import SQSingleQueue
from ..sq_event import SQEventManager
class SQFactory:
    def __init__(self, helper_factory: SQHelperFactory = None,
                 plugin_list: Union[List[str], None] = None):
        if plugin_list is None:
            self.plugin_list = []
        else:
            self.plugin_list = plugin_list
        self.factory_map: dict[str, Callable[..., SQObject]] = {
            'SQClock': SQClock,
            'SQFilter': SQFilter,
            'SQMerger': SQMerger,
            'SQMux': SQMux,
            'SQDemux': SQDemux,
            'SQPacketGenerator': SQPacketGenerator,
            'SQPktProcessor': SQPktProcessor,
            'SQSimulator': SQSimulator,
            'SQSplitter': SQSplitter,
            'SQPktSink': SQPktSink,
            'SQQueue': SQSingleQueue,
        }
        self.event_mgr = SQEventManager()
        self.helper_factory = helper_factory
        if self.helper_factory is None:
            self.helper_factory = SQHelperFactory()
        self.plugin_loader = SQPluginLoader(self.helper_factory)
        for plugin in self.plugin_list:
            self.load_plugin(plugin)

    def load_plugin(self, plugin_name: str) -> None:
        self.plugin_loader.load_plugin(plugin_name)

    def create(self, obj_type: str, data: dict[str, Any]) -> SQObject:
        data['helper_factory'] = self.helper_factory
        data['event_mgr'] = self.event_mgr
        return self.factory_map[obj_type](data)



from typing import List, Union
from ..sq_event import SQEvent
from .sq_event_queue import SQEventQueue
from ..sq_logger import SQLogger
from ..sq_time_base import SQTimeBase


class SQEventManager:
    def __init__(self, name: str = ""):
        self.name = name
        self.event_queue_list: List[SQEventQueue] = []
        self.schedule_queue: List[SQEvent] = []
        self.logger = SQLogger(self.__class__.__name__, self.name)

    def set_log_level(self, level):
        self.logger.set_level(level)

    def schedule(self, _event: SQEvent, when: int):
        if _event.has_handlers():
            self.logger.debug(f'Schedule Event {_event.name} after {when} sim time')
            _event.scheduled_tick = when + SQTimeBase.get_current_sim_time()
            self.schedule_queue.append(_event)
        else:
            self.logger.debug(f'Event {_event.name} has no handlers')

    def get_event_queue(self, index: int) -> Union[SQEventQueue, None]:
        queue: Union[SQEventQueue, None] = None
        if self.event_queue_list is not None:
            if len(self.event_queue_list) > index:
                queue = self.event_queue_list[index]
            else:
                queue = SQEventQueue()
                self.event_queue_list.append(queue)
        return queue

    def set_schedule_queue(self):
        for _event in self.schedule_queue:
            queue = self.get_event_queue(_event.owner.evt_q)
            if queue is not None:
                queue.schedule(_event=_event)
        self.schedule_queue.clear()

    def run(self):
        self.logger.debug(f'SQEventManager: run')
        self.set_schedule_queue()
        for queue in self.event_queue_list:
            while queue.get_next_event() is not None:
                next_evt = queue.pop_next_event()
                self.logger.debug(f'SQEventManager: Handling Event {next_evt.name}')
                for action in next_evt.actions:
                    action(next_evt)



from abc import ABC
from typing import List, Union

from .sq_event import SQEvent, SQEventManager, EventType
from .sq_logger import SQLogger
from .sq_statistics import SQStatistics
from .sq_packet import SQMetadata
from .sq_plugin import SQHelper


class SQObject(ABC):
    """
    Base class for all objects in the simulation
    """

    def __init__(self, data: dict[str, any]):

        self.name = data.get('name', "")
        self.helper_factory= data.get('helper_factory', None)
        if self.helper_factory is None:
            raise ValueError('helper_factory should be provided')
        self.helper_name = data.get('helper', "default")
        self.helper: Union[SQHelper, None] = None
        data['owner'] = self
        self.helper = self.helper_factory.create(name=self.helper_name, data=data)
        if self.name == "":
            raise ValueError('name should be provided')
        self.evt_q: int = data.get('event_q', 0)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.event_manager = data.get('event_mgr', None)
        if self.event_manager is None:
            raise ValueError('event_mgr should be provided')
        self.tick: int = 0
        self.self_starting = False
        self.children = data.get('children', [])
        self.statistics = SQStatistics()
        self.statistics_properties = []
        self.is_self_ticking: bool = data.get('is_self_ticking', False)
        self.data_flow_map = []
        self.tick_evt = SQEvent(_name=f'{self.name}_tick',
                                owner=self)
        self.tick_evt.type = EventType.TICK_EVENT

        self.start_evt = SQEvent(_name=f'{self.name}_start',
                                 owner=self)
        self.start_evt.type = EventType.START_EVT
        self.start_evt.add_handler(self.process_packet)
        self.finish_evt = SQEvent(_name=f'{self.name}_finish',
                                  owner=self)
        self.finish_evt.type = EventType.PROCESS_EVT

        self.metadata_evt = SQEvent(_name=f'{self.name}_metadata',
                                    owner=self)

        self.metadata_evt.type = EventType.METADATA_EVT

    def __repr__(self):
        return self.name

    def set_log_level(self, level):
        self.logger.set_level(level)

    def subscribe_metadata(self, owner: str, data_name: str):
        if owner is self.name:
            self.logger.error(f"Cannot subscribe to metadata from other objects")
            return
        else:
            self.data_flow_map.append(SQMetadata(owner=owner, name=data_name, value=None))

    def _is_metadata_subscribed(self, owner: str, data_name: str):
        for data in self.data_flow_map:
            if data.owner == owner and data.name == data_name:
                return True
        return False

    def get_metadata_received(self, owner: str, data_name: str):
        for data in self.data_flow_map:
            if data.owner == owner and data.name == data_name:
                return data.value
        return None

    def update_subscribed_metadata(self, metadata: SQMetadata):
        for data in self.data_flow_map:
            if data.owner == metadata.owner and data.name == metadata.name:
                data.value = metadata.value

    def register_property(self, name: str, owner=None):
        if owner is not None:
            self.statistics_properties.append({'owner': owner, 'name': name})
        else:
            self.statistics_properties.append({'owner': self, 'name': name})

    def process_data(self, evt: SQEvent):
        self.logger.info(f'Process Metadata {evt.owner.name}::{evt.data.name} on Tick {self.tick}')

    def _process_metadata(self, evt: SQEvent):
        if evt.type == EventType.METADATA_EVT:
            if evt.data is not None:
                if isinstance(evt.data, SQMetadata):
                    if self._is_metadata_subscribed(evt.data.owner, evt.data.name):
                        self.update_subscribed_metadata(evt.data)
                        self.process_data(evt)
                else:
                    self.logger.error(f"Metadata {evt.data} not a SQMetadata object")
        else:
            self.logger.error(f"Ignoring Metadata Event {evt.name} from {evt.owner.name}")

    def init(self):
        self.logger.info(f'init')
        for child in self.children:
            assert isinstance(child, SQObject), "all child objects of an SQObject must be derived from SQObject"
            child.init()

    def deinit(self):
        self.logger.info(f'deinit')
        for child in self.children:
            child.deinit()

    def read_statistics(self):
        return self.statistics

    def start(self):
        self.logger.info(f'start')
        if not self.self_starting:
            self.event_manager.schedule(self.start_evt, when=1)
        if self.is_self_ticking:
            self.logger.info(f'is self ticking')
            self.self_connect()
        self.tick = 0
        for child in self.children:
            child.start()

    def finish_indication(self, data=None, when=1):
        self.logger.info(f'finish indication')
        self.finish_evt.data = data
        self.event_manager.schedule(self.finish_evt, when=when)

    def data_indication(self, data: SQMetadata, when=1):
        self.logger.info(f'data indication')
        self.metadata_evt.data = data
        self.event_manager.schedule(self.metadata_evt, when=when)

    def self_trigger(self, when=1):
        self.logger.info(f'self_trigger at {self.tick}')
        self.event_manager.schedule(self.tick_evt, when=when)

    def collect_statistics(self):
        for p in self.statistics_properties:
            self.statistics.add(p['name'], getattr(p['owner'], p['name']), self.name)

    def process_packet(self, evt: SQEvent):
        self.logger.info(f'Process Event {evt.owner.name}::{evt.name} on Tick {self.tick}')
        if self.is_self_ticking:
            self.self_trigger()
        self.collect_statistics()

    def control_flow(self, obj: "SQObject", **kwargs):
        self.finish_evt.add_handler(obj.process_packet)
        return obj

    def data_flow(self, obj: "SQObject", metadata: List[str]):
        self.metadata_evt.add_handler(obj._process_metadata)
        for data in metadata:
            obj.subscribe_metadata(owner=self.name, data_name=data)
        return obj

    def self_connect(self):
        self.tick_evt.add_handler(self.process_packet)

    def disconnect_control_flow(self, obj: "SQObject"):
        self.logger.info(f"disconnecting the observer{obj.name} from control flow")
        self.finish_evt.remove_handler(obj.process_packet)

    def disconnect_data_flow(self, obj: "SQObject"):
        self.logger.info(f"disconnecting the observer{obj.name} from dataflow")
        self.metadata_evt.remove_handler(obj._process_metadata)

    def get_current_tick(self):
        return self.tick

    def update_tick(self, ticks: int):
        self.tick += ticks
        self.logger.info(f'Updating Current Tick {self.tick}')



from ..sq_object import SQObject
from ..sq_logger import SQLogger
from ..sq_event import SQEvent
from ..sq_event.sq_event_manager import SQEventManager
from ..sq_time_base import SQTimeBase


class SQClock(SQObject):
    """
    A Clock object that ticks at a given frequency based on the clk_divider
    """

    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.clk_divider = data.get('clk_divider', 1)
        self.is_self_ticking = True

    def process_packet(self, evt: SQEvent):
        current_sim_time = SQTimeBase.get_current_sim_time()
        # only tick if the current sim time is a multiple of clk_divider
        if current_sim_time % self.clk_divider == 0:
            self.tick += 1
            self.logger.info(
                f" Clock Tick = {self.tick} on sim time {current_sim_time}")
            self.finish_indication()
        else:
            self.logger.debug(
                f" Skipping at Clock Tick = {self.tick} on sim time {current_sim_time}")
        super().process_packet(evt)



from typing import Union

from ..sq_object import SQObject
from ..sq_event import SQEventManager, SQEvent
from ..sq_logger import SQLogger
from .states import SQPktGeneratorState
from ..sq_queue import SQQueue
from ..sq_clock import SQClock
from .states import SQGeneratorStateFactory
from ..sq_plugin import SQHelper


class SQPacketGenerator(SQObject):
    """
    Base class for all Packet Generators in the simulation
    The class implements the basic functionality of a packet generator
    """

    def __init__(self, data: dict[str, any]):

        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        factory = SQGeneratorStateFactory()
        self.state: SQPktGeneratorState = factory.create_state("GENERATING", self)
        self.generated_pkts = 0
        self.total_pkts = 0
        self.output_q = data.get('output_q', None)
        if self.output_q is None:
            raise ValueError('output_q should be provided')
        self.clk = data.get('clk', None)
        if self.clk is not None:
            self.clk.control_flow(self)
        else:
            raise ValueError('Clock not provided')
        if self.helper is None:
            raise ValueError('Helper not provided')

        self.register_property('generated_pkts')
        self.register_property('total_pkts')
        self.packets = []

    def set_state(self, state):
        self.state = state

    def process_packet(self, evt: SQEvent):
        super().process_packet(evt)
        if evt.owner is self.clk:
            self.state.process_packet(evt)
        else:
            if evt.name != f'{self.name}_start':
                self.logger.error(f'Ignoring Events other than Clock Events {evt}')



from ...sq_packet import SQPacket
from dataclasses import dataclass


@dataclass
class SQGenericPacket(SQPacket):
    id: int = 0
    priority: int = 0
    class_name: str = ""
    src: str = ""
    destination: str = ""



from ..sq_object import SQObject
from ..sq_clock import SQClock
from ..sq_queue import SQQueue
from ..sq_event import SQEvent
from ..sq_logger import SQLogger


class SQPktSink(SQObject):
    def __init__(self, data: dict[str, any]):
        super().__init__(data)
        self.logger = SQLogger(self.__class__.__name__, self.name)
        self.input_q = data.get('input_q', None)
        if self.input_q is None:
            raise ValueError('Input Queue should be provided')
        self.clk = data.get('clk', None)
        if self.clk is not None:
            self.clk.control_flow(self)
        else:
            raise ValueError('Clock should be provided')

    def process_packet(self, evt: SQEvent):
        super().process_packet(evt)
        if evt.owner is self.clk:
            curr_pkt = self.input_q.pop()
            self.tick += 1
            if curr_pkt is not None:
                self.logger.info(f' Terminated the Packet {curr_pkt} ')
            else:
                self.logger.info(f' No Packet to Terminate')
        else:
            if evt.name != f'{self.name}_start':
                self.logger.error(f'Ignoring Events other than Clock Events {evt}')



from .sq_generator_state import SQPktGeneratorState
from ...sq_event import SQEvent


class SQPktGeneratorGenState(SQPktGeneratorState):
    def process_packet(self, evt: SQEvent):
        self.owner.generated_pkts = 0
        self.owner.logger.info(f'Generating Packets')
        self.owner.packets = [d for d in self.owner.helper.generate_packets()][0]
        self.owner.logger.info(f'Generated {len(self.owner.packets)} Packets')
        self.owner.set_state(self.factory.create_state(name='QUEUING', owner=self.owner))

    def get_state_name(self):
        return f'GENERATING'
